import re
from typing import Tuple

from nbox.relics.base import BaseStore

class S3(BaseStore):
  def __init__(self, s3_url):
    import boto3
    self._boto3 = boto3 # alias
    self.client = self._boto3.client('s3')
    self.s3_url = s3_url
    bucket, prefix = self._split_to_bucket_prefix(s3_url)

  def _split_to_bucket_prefix(self, s3_url: str) -> Tuple[str, str]:
    # clean the url
    if not s3_url.startswith("http"):
      s3_url = re.sub(r"^http[s]://", "", s3_url)
    if not s3_url.startswith("s3."):
      raise Exception(f"Invalid S3 URL: {s3_url}")
    return "", "/path"    

  def _put(self, key: str, value: bytes, ow: bool = False) -> None:
    self.client.put_object(
      Bucket=self.s3_url.split('/')[2],
      Key=key,
      Body=value,
      ACL='public-read',
      StorageClass='STANDARD'
    )



import os
import dill
from pathlib import Path
from typing import Any, Union

from nbox.utils import logger, FileLogger, env
from nbox.relics.base import BaseStore

class LocalStore(BaseStore):
  """
  Cache structure is like this:

  {cache_dir}/
    items/
      fdedc958b417adf63278938efa53c3b381f576446aede80bbc8f2c05320fcb4b
      1459d663d14b7b7ad82ebe5c98c8a0397b21d4e2b9f4711562746a5cb48f86c4
      ...
    _objects.bin # contains the entirity of information on this cache
    activity.log # contains the logs of this cache
  """
  def __init__(self, cache_dir: str = None, create: bool = True) -> Union[None, Any]:
    self.cache_dir = cache_dir or os.path.join(env.NBOX_HOME_DIR(), "relics")
    logger.info(f"Connecting object store: {self.cache_dir}")
    self._objects = {} # <key: item_path>
    self._objects_bin_path = f"{self.cache_dir}/_objects.bin"
    self._file_logger_path = f"{self.cache_dir}/activity.log"

    # format = "pickle/pyarrow/tfrecords" in _put means have to write in the _get

    # Create the cache directory if it doesn't exist
    if not os.path.exists(self.cache_dir):
      Path(self.cache_dir).mkdir(parents=create)
      # open(self._objects_bin_path, "w").close()
      # open(self._file_logger_path, "w").close()
    
    # load the data from the cache
    if os.path.isdir(self.cache_dir):
      if not os.path.exists(self._objects_bin_path):
        self._objects = {}
      else:
        with open(self._objects_bin_path, "rb") as f:
          self._objects = dill.load(f)
      if not os.path.exists(f"{self.cache_dir}/items"):
        os.makedirs(f"{self.cache_dir}/items")
      self._logs = FileLogger(self._file_logger_path)
    else:
      raise Exception(f"{self.cache_dir} is not a directory")

  def _read_state(self):
    if os.path.exists(self._objects_bin_path):
      with open(self._objects_bin_path, "rb") as f:
        self._objects = dill.load(f)

  def _write_state(self):
    with open(self._objects_bin_path, "wb") as f:
      dill.dump(self._objects, f)
  
  def _put(self, key: str, value: bytes, ow: bool = False,) -> None:
    self._read_state()

    fp = f"{self.cache_dir}/{self._clean_key(key)}"
    _key = self.get_id(fp)
    item_path = f"{self.cache_dir}/items/{_key}"
    
    if os.path.exists(item_path) and not ow:
      raise Exception(f"File already exists: {fp}")
    with open(item_path, "wb") as f:
      dill.dump(value, f)

    # update the statae
    self._objects[fp] = item_path
    self._logs.info(f"PUT {key}")
    self._write_state()

  def _get(self, key: str) -> bytes:
    self._read_state()

    fp = f"{self.cache_dir}/{self._clean_key(key)}"
    item_path = self._objects.get(fp, None)
    if item_path is not None and os.path.exists(item_path):
      self._logs.info(f"GET {key}")
      with open(item_path, "rb") as f:
        out = dill.load(f)
      return out
    return None

  def _delete(self, key: str) -> None:
    self._read_state()

    fp = f"{self.cache_dir}/{self._clean_key(key)}"
    item_path = self._objects.get(fp, None)
    if item_path is not None and os.path.exists(item_path):
      os.remove(item_path)
      del self._objects[fp]
      self._logs.warning(f"DELETE {key}")
      self._write_state()



import re
from uuid import uuid4
from hashlib import sha256
from typing import Any, Tuple

from abc import abstractmethod

class BaseStore(object):
  # Typical OOPs things

  def _get_uuid(self, s = True) -> str:
    if s:
      return str(uuid4()).split('-')[0]
    return str(uuid4())

  def get_id(self, key: str, random: bool = False) -> str:
    _key = sha256(key.encode('utf-8')).hexdigest()
    if random:
      return str(uuid4()).split('-')[0] + "-" + _key
    return _key

  def _clean_key(self, key: str) -> str:
    return key.strip('/')

  @abstractmethod
  def _put(self, key: str, value: Any) -> None:
    pass

  @abstractmethod
  def _get(self, key: str) -> Any:
    pass

  @abstractmethod
  def _delete(self, key: str) -> None:
    pass

  def put(self, key: str, value: Any, ow: bool = False) -> None:
    self._put(key, value, ow)

  def get(self, key: str, default = None) -> Any:
    out = self._get(key)
    if out is None:
      return default
    return out

  def delete(self, key: str) -> None:
    self._delete(key)

  def __getitem__(self, key: str):
    return self.get(key)



"""
NimbleBox' object store that can connect to your backend of choice (AWS S3, etc.) to provide persistant storage beyond instances.
"""

import re
from enum import Enum
from time import sleep

from nbox.init import nbox_ws_v1
from nbox.relics.local import LocalStore

class RelicTypes(Enum):
  """
  Types of relics.
  """
  UNSET = 0
  LOCAL = 1 # all the files in this mode are going to be stored in some directory
  NBX = 2 # this mode means that relic is the one run by NimbleBox
  AWS_S3 = 3 # AWS S3
  GCP_BUCKET = 4 # Google Cloud Storage
  AZURE_BLOB = 5 # Azure Blob Storage
  LOCAL_REDIS = 6 # store and retrieve data from local redis


class Relic():
  _mode = RelicTypes.UNSET
  def __init__(self, id_or_url, workspace_id = "personal", **kwargs):
    """Connect to your Relic of choice."""
    # go over the bunch of cases that id_or_url can be and set the mode
    if id_or_url.startswith('http'):
      out = re.findall(r".ai\/(\w+)\/relics\/(.*)$", id_or_url)
      if not out:
        raise Exception(f"Invalid URL: '{id_or_url}'")
      self.workspace_id = out[0][0]
      self.relic_id = out[0][1]
      self.url = id_or_url

    elif id_or_url.startswith('local:'):
      cache_dir = id_or_url[6:]
      if cache_dir == "":
        cache_dir = None
      self.store = LocalStore(cache_dir, **kwargs)
      self._mode = RelicTypes.LOCAL # change the mode to local

    elif re.findall(r"^\w{8}$", id_or_url):
      self.workspace_id = workspace_id
      self.relic_id = id_or_url

    # create the subway stub
    # self.stub = nbox_ws_v1.u(workspace_id).relics.u(self.relic_id)

    # hit NBX APIs to get more information about this
    # self.refresh()

  def refresh(self):
    data = self.stub()
    
    builder = data["builder"] # some information about the builder like cloud vendor, etc
    api_meta = data["vendor_metdata"] # say things that I get in get_upload_url()
    file_meta = data["file_meta"] # folder also are just file references

    self.name = file_meta["name"]
    self.is_file = file_meta["is_file"]
    self.size = file_meta["size"]

  @property
  def metadata(self):
    return {}

  def download(self, key, local_path):
    """download and store the file to local_path"""
    pass

  def put(self, obj, key, ow):
    """put the object to NBX"""
    self.store.put(obj, key, ow)

  def get(self, key):
    """get the file from NBX"""
    out = self.store.get(key)
    while out == None:
      sleep(1)
      out = self.store.get(key)
    return out

  def has(self, key):
    """check if the file exists in NBX"""
    return self.store.get(key) != None

  def upload(self, local_path, key = None):
    """upload the file to NBX"""
    pass

  def delete(self, key):
    """delete the file from NBX"""
    self.store.delete(key)
